/////////////////////////////////////////////////////
// VP Digitale Signalprozessoren
/////////
// Auernigg Franz-Josef
// 0120133
//////
// Filter: Biorthogonal with Lifting
/////////////////////////////////////////////////////

#include "filter.h"
#include "math.h"
//#include <stdio.h>

const float k_a = -1.586134342,
			k_b = -0.052980118,
			k_c =  0.882911075,
			k_d =  0.443506852,
			k_e =  1.230174105;

const floatsize = sizeof(float) ;

#define select(r0,r1,r2,r3)		(unsigned int)((r0%2)+(r0/2)*2+(r1%2)*4+(r1/2)*8+(r2%2)*16+(r2/2)*32+(r3%2)*64+(r3/2)*128)


/////////////////////////////////////////////////////
// Filter: Biorthogonal with Lifting , without sse
/////////////////////////////////////////////////////
void filter_bio79lift_plain (float *source, float *dest, int size)
{
  float o,p,q,r,s,t,u,v,w;
  float inv_k_e = 1 / k_e ;
  float *L, *H ;
  int n= 0 ;
  float *X ;

  X = source ;
  L = dest ;
  H = & dest[size/2] ;

  o = q = X[2*n+2] ;
  s = 0 ;
  u = 0 ;
  w = 0 ;


  for (n = 0 ; n < size/2 - 4; n++ )
  {
	  // The Bioorthogonal filter
	  /*
	  X(2n+1) = X(2n+1) + a * (X(2n) + X(2n+2))
	  X(2n) = X(2n) + b * (X(2n-1) + X(2n+1)

	  X(2n+1) = X(2n+1) + c * (X(2n) + X(2n+2))
	  X(2n) = X(2n) + d * (X(2n-1) + X(2n+1) )

	  X(2n+1) = -e * X(2n+1)
	  X(2n) = (1/e) * X(2n)
	  */

	  /////////////////////////////////////////////////////
	  // -> Implemented as:
	  o = q ;
	  p = X[ 2*n+3 ] ;
	  q = X[ 2*n+4 ] ;
	  r = s ;
	  s = p + k_a * ( o + q ) ;
	  t = u ;
	  u = o + k_b * ( r + s ) ;
	  v = w ;
	  w = r + k_c * ( t + u ) ;
	  L[n] = ( t + k_d * ( v + w ) )  * inv_k_e ;
	  H[n] =  w * (0-k_e) ;
  }

}
/////////////////////////////////////////////////////

/////////////////////////////////////////////////////
// Filter: Biorthogonal with Lifting , sse
/////////////////////////////////////////////////////
void filter_bio79lift_vec (float *source, float *dest, int size)
{
  typedef int v4sf __attribute__ ((mode(V4SF)));

  float __attribute__ ((aligned (16))) fop_a[] = { k_a, k_a, k_a, k_a };
  float __attribute__ ((aligned (16))) fop_b[] = { k_b, k_b, k_b, k_b };
  float __attribute__ ((aligned (16))) fop_c[] = { k_c, k_c, k_c, k_c };
  float __attribute__ ((aligned (16))) fop_d[] = { k_d, k_d, k_d, k_d };
  float __attribute__ ((aligned (16))) fop_e[] = { -k_e, -k_e, -k_e, -k_e };
  float __attribute__ ((aligned (16))) inv_fop_e[] = { 1/k_e, 1/k_e, 1/k_e, 1/k_e };
  float __attribute__ ((aligned (16))) fop_zero[] = { 0,0,0,0 };
  
  v4sf a, b ;
  v4sf	op_a = * (v4sf *) &fop_a,
		op_b = * (v4sf *) &fop_b,
		op_c = * (v4sf *) &fop_c,
		op_d = * (v4sf *) &fop_d,
		op_e = * (v4sf *) &fop_e,
		inv_op_e = * (v4sf *) &inv_fop_e,
		o, p, q, r, s=*(v4sf*)&fop_zero, t, u=*(v4sf*)&fop_zero, v, w=*(v4sf*)&fop_zero,
		helper;

  float *sp, *L, *H ;
  int i ;

  sp = source ;
  L = dest ;
  H = & dest[size/2] ;

  if ( ((int) source % 16) || ((int) dest % 16) )
  {
     printf("data not aligned!  ");
     return;
  }

  /////////////////////////////////////////////////////
  // This is the invariant calculation shemata for algorithmen presented in the "while loop"
  //
  // Data:..2n-1	2n	  2n+3	2n+4	...
  //         ||		||		||	 ||
  // .	.	.		o		p	 q
  //	\	|	/	|	\	|	/
  //		a 		|		a
  //  		| \ =s	|	   /
  //   b	|   ---	b  --- 
  //	\	| =t  /	| =u
  //	 \	| 	 /	|
  //		c -- 	
  //		|
  // 
  //		..........
  // The Formula for the Step 1 is: result_a = odd + (even_left+even_right)*a
  // In Step 2 the	odd = even_left
  //				even_right = result_a (result of last step)
  //				even_left  = s
  // This algorithm is the same for every step except the last one. 
  // Only the variables t,s,... and the coefficients vary.
  //
  /////////////////////////////////////////////////////

  while( sp < (source + size-4) )
  {
	  /////////////////////////////////////////////////////
	  // load 8 floats of input data

	  // - we need c,d because o + q (floats) is implemented as a + c (v4sf-regs)
	  a = __builtin_ia32_loadaps (sp+4) ;
	  b = __builtin_ia32_loadaps (sp+8) ;
	  
	  // split data into even and odd (of index number)
	  // - even data - right side
	  q = __builtin_ia32_shufps (a, b, select(0,2,0,2) );

	  /* This Code has been not optimized but is easier to understand
	  // loadups because with loadaps input data must be aligned in 16 Byte units
	  a = __builtin_ia32_loadups (sp+2) ;
	  b = __builtin_ia32_loadups (sp+6) ;

	  // split data into even and odd (of index number)
	  // - even data - left side
	  o = __builtin_ia32_shufps (a, b, select(0,2,0,2) );
	  // - odd  data
	  p = __builtin_ia32_shufps (a, b, select(1,3,1,3) ) ;*/

	  // optimized:
	  // loadups because with loadaps input data must be aligned in 16 Byte units
	  helper = __builtin_ia32_loadaps (sp) ;
	  helper = __builtin_ia32_shufps (helper, a, select(2,3,0,1) ) ;
	  b = __builtin_ia32_shufps (a, b, select(2,3,0,1) ) ;
	  a = helper ;

	  // split data into even and odd (of index number)
	  // - even data - left side
	  o = __builtin_ia32_shufps (a, b, select(0,2,0,2) );
	  // - odd  data
	  p = __builtin_ia32_shufps (a, b, select(1,3,1,3) ) ;

	  sp += 8 ;

	  
	  /////////////////////////////////////////////////////

	  /////////////////////////////////////////////////////
	  // Step 1:
	  /////////////////////////////////////////////////////
	  
	  // save result s; it will be used in the next iteration step
	  r = s ;

	  // Calculate the equation
	  s = ( o +  q ) *  op_a + p ;

	  /////////////////////////////////////////////////////

	  /////////////////////////////////////////////////////
	  // Step 2:
	  /////////////////////////////////////////////////////

	  helper = __builtin_ia32_shufps ( r, s, select(3,3,0,1) ) ;
	  r = __builtin_ia32_shufps ( helper, s , select(1,2,1,2) ) ;

	  // save last result of u
	  t = u ;

	  // Calculate the equation
	  u = ( r +  s ) *  op_b + o  ;
	  
	  /////////////////////////////////////////////////////

	  /////////////////////////////////////////////////////
	  // Step 3: (similar to Step 2)
	  /////////////////////////////////////////////////////

	  helper = __builtin_ia32_shufps ( t, u, select(3,3,0,1) ) ;
	  t = __builtin_ia32_shufps ( helper, u , select(1,2,1,2) ) ;

	  // save last result
	  v = w ;

	  // Calculate the equation
	  w = ( t +  u ) *  op_c + r ;
	  
	  /////////////////////////////////////////////////////

	  /////////////////////////////////////////////////////
	  // Step 4+5: (similar to Step 2)
	  /////////////////////////////////////////////////////

	  helper = __builtin_ia32_shufps ( v, w, select(3,3,0,1) ) ;
	  v = __builtin_ia32_shufps ( helper, w , select(1,2,1,2) ) ;

	  // Calculate the equation
	  a = ( ( v + w ) * op_d + t ) *  inv_op_e  ;
	  b = w * op_e ;

	  // - Save Data
	  __builtin_ia32_storeaps ( L , a );
	  __builtin_ia32_storeaps ( H , b );

	  L += 4 ;
	  H += 4 ;

	  /////////////////////////////////////////////////////
  }
}
/////////////////////////////////////////////////////

void rfilter_bio79lift_plain (float *source, float *dest, int size)
{
/*
  float r = M_SQRT2 * .5;
  float a, b;
  float *spl, *sph, *dp;
  int i;

  spl = source;
  sph = & source[size/2];
  dp  = dest;
*/
  /*for (i = size / 2; i > 0; i --)
  {
      a = *spl * r;
      spl ++;
      b = *sph * r;
      sph ++;
      *dp = a + b;
      dp ++;
      *dp = a - b;
      dp ++;
  }*/
}

void rfilter_bio79lift_vec (float *source, float *dest, int size)
{/*
  typedef int v4sf __attribute__ ((mode(V4SF)));

  int i;
  float r = M_SQRT2 * .5;
  float __attribute__ ((aligned (16))) fop[] = {r, r, r, r};
  v4sf op;
  v4sf a, b, c, d, e, f;
  float *spl, *sph, *dp;

  op = * (v4sf *) &fop;
  spl = source;
  sph = & source[size/2];
  dp  = dest;
*/
  /*for (i = size / 8; i > 0; i --)
  {
      a = __builtin_ia32_loadaps (spl) * op;
      spl += 4;
      b = __builtin_ia32_loadaps (sph) * op;
      sph += 4;
      c = a + b;
      d = a - b;
      e = __builtin_ia32_unpcklps (c, d);
      f = __builtin_ia32_unpckhps (c, d);
      __builtin_ia32_storeaps (dp, e);
      dp += 4;
      __builtin_ia32_storeaps (dp, f);
      dp += 4;
  }*/
}
