#include "filter.h"
#include "math.h"
//#include <stdio.h>

void filter_haar_plain (float *source, float *dest, int size)
{
  float r = M_SQRT2 * .5;
  float a, b;
  float *sp, *dpl, *dph;
  int i;

  sp = source;
  dpl = dest;
  dph = & dest[size/2];

  for (i = size / 2; i > 0; i --)
  {
      a = *sp * r;
      sp ++;
      b = *sp * r;
      sp ++;
      *dpl = a + b;
      dpl ++;
      *dph = a - b;
      dph ++;
  }
}

void filter_haar_vec (float *source, float *dest, int size)
{
  typedef int v4sf __attribute__ ((mode(V4SF)));

  int i;
  float r = M_SQRT2 * .5;
  float __attribute__ ((aligned (16))) fop[] = {r, r, r, r};
  v4sf op;
  v4sf a, b, c, d;
  float *sp, *dpl, *dph;

  /*
  if (((int) source % 16) || ((int) dest % 16))
  {
     fputs ("data not aligned!", stderr);
     return;
  }
  */

  op = * (v4sf *) &fop;
  sp = source;
  dpl = dest;
  dph = & dest[size/2];

  for (i = size / 8; i > 0; i --)
  {
      a = __builtin_ia32_loadaps (sp) * op;
      sp += 4;
      b = __builtin_ia32_loadaps (sp) * op;
      sp += 4;
      c = __builtin_ia32_shufps (a, b, 0x88);
      d = __builtin_ia32_shufps (a, b, 0xdd);
      __builtin_ia32_storeaps (dpl, c + d);
      __builtin_ia32_storeaps (dph, c - d);
      dpl += 4;
      dph += 4;
  }
}

void rfilter_haar_plain (float *source, float *dest, int size)
{
  float r = M_SQRT2 * .5;
  float a, b;
  float *spl, *sph, *dp;
  int i;

  spl = source;
  sph = & source[size/2];
  dp  = dest;

  for (i = size / 2; i > 0; i --)
  {
      a = *spl * r;
      spl ++;
      b = *sph * r;
      sph ++;
      *dp = a + b;
      dp ++;
      *dp = a - b;
      dp ++;
  }
}

void rfilter_haar_vec (float *source, float *dest, int size)
{
  typedef int v4sf __attribute__ ((mode(V4SF)));

  int i;
  float r = M_SQRT2 * .5;
  float __attribute__ ((aligned (16))) fop[] = {r, r, r, r};
  v4sf op;
  v4sf a, b, c, d, e, f;
  float *spl, *sph, *dp;

  op = * (v4sf *) &fop;
  spl = source;
  sph = & source[size/2];
  dp  = dest;

  for (i = size / 8; i > 0; i --)
  {
      a = __builtin_ia32_loadaps (spl) * op;
      spl += 4;
      b = __builtin_ia32_loadaps (sph) * op;
      sph += 4;
      c = a + b;
      d = a - b;
      e = __builtin_ia32_unpcklps (c, d);
      f = __builtin_ia32_unpckhps (c, d);
      __builtin_ia32_storeaps (dp, e);
      dp += 4;
      __builtin_ia32_storeaps (dp, f);
      dp += 4;
  }
}
