#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "filter.h"

#define FUNJOIN(p1, p2, p3) FUNJOIN2(p1, p2, p3)
#define FUNJOIN2(p1, p2, p3) p1 ## p2 ## p3

#define FILTTYPE haar
//#define FILTTYPE bio79
//#define FILTTYPE bio79lift
//#define FILTTYPE daub8
//#define FILTTYPE bio412

int main ()
{
  int size = 16 * 1048576;
  int i;
  clock_t t1, t2, t3;
  clock_t pms = CLOCKS_PER_SEC / 1000;

  float *msource    = (float *) malloc (size * sizeof (float) + 100);
  float *source = msource + 50;
  float *destplain = (float *) malloc (size * sizeof (float));
  float *mdestvec   = (float *) malloc (size * sizeof (float) + 100);
  float *destvec = mdestvec + 50;
  float *mdest2plain = (float *) malloc (size * sizeof (float) + 100);
  float *dest2plain = mdest2plain + 50;
  float *mdest2vec   = (float *) malloc (size * sizeof (float) + 100);
  float *dest2vec = mdest2vec + 50;
  double error, diff, verror;

  while ((int) source % 16) source ++;
  while ((int) destvec % 16) destvec ++;
  while ((int) dest2vec % 16) dest2vec ++;

  for (i = 0; i < size; i ++)
      source[i] = i * 0.1;
  for (i = 0; i < 100; i ++)
  {
      msource[i] = 0.0;
      msource[size + 99 - i] = 0.0;
  }

  t1 = clock ();


  FUNJOIN (filter_, FILTTYPE, _plain) (source, destplain, size);

  t2 = clock ();

  FUNJOIN (filter_, FILTTYPE, _vec) (source, destvec  , size);

  t3 = clock ();


  error = 0.0;
  for (i = 0; i < size; i ++)
  {
      diff = destplain[i] - destvec[i];
      error += diff * diff;
  }
  error /= (double) size;

  printf ( "forward:\n plain time: %d ms\n vector time: %d ms\n speedup: %f\n error: %f\n"
         , (t2 - t1) / pms
         , (t3 - t2) / pms
         , (double) (t2 - t1) / (double) (t3 - t2)
         , error
         );

  t1 = clock ();

  FUNJOIN (rfilter_, FILTTYPE, _plain) (destplain, dest2plain, size);

  t2 = clock ();

  FUNJOIN (rfilter_, FILTTYPE, _vec) (destvec, dest2vec, size);

  t3 = clock ();

  error = 0.0;
  verror = 0.0;
  for (i = 0; i < size; i ++)
  {
      diff = dest2plain[i] - source[i];
      error += diff * diff;
      diff = dest2vec[i] - source[i];
      verror += diff * diff;
  }
  error /= (double) size;
  verror /= (double) size;

  printf ( "reverse:\n plain time: %d ms\n vector time: %d ms\n speedup: %f\n plainerror: %f\n vectorerror: %f\n"
         , (t2 - t1) / pms
         , (t3 - t2) / pms
         , (double) (t2 - t1) / (double) (t3 - t2)
         , error
         , verror
         );

  free (msource);
  free (destplain);
  free (mdestvec);
  free (mdest2plain);
  free (mdest2vec);
}
